#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef LINUX_TEST
#include <ogcsys.h>
#include <gccore.h>
#include <gctypes.h>
#include "ogc/ipc.h"
#else
#ifdef CYGWIN
#include <ncurses/curses.h>
#else
#include <curses.h>
#endif
#endif

#include "os_menu.h"
#include <string.h>

#ifndef LINUX_TEST
static void *xfb = NULL;
static GXRModeObj *rmode = NULL;
#endif

#define BLACK_COLOR 0x00800080

#define NULL_STROKE 0x00
#define RETURN_KEY  0x0D
#define MAX_LINE_LENGTH 255

#define UP_KEY 0x55
#define DOWN_KEY 0x53

#define KEY_JUMP_R 101
#define KEY_JUMP_L 113

#define MENU_SCREEN_SIZE 11
#define LOG_SCREEN_SIZE 7
#define MAX_PAYLOAD_SIZE 20

#define CANCEL 0xFFFFFFFF

#define MODE_CHANGE 999

long synch_time = 0;

char payload[100];
char sLog[100];
char tfoot[50];
char tResult[30];
char current_header[200];
char buffer[0x100];  
char tstr[5];

char* version = "0.1.2";

#define HEX_MODE 0
#define BINARY_MODE 1

char* input_mode[2] = {"HEX","BINARY"};
int current_input_mode = BINARY_MODE;

_log_type my_log;
int my_kbd;
unsigned char key_press;

struct _ioctlv_data_type{
   int ioctl;
   int cnt_in;
   int cnt_io;
   unsigned char bmRequestType;
   unsigned char bmRequest;
   unsigned short wValue;
   unsigned short wIndex;
   unsigned char data[20];
   int len;
};

typedef struct _ioctlv_data_type _ioctlv_data_type;

_ioctlv_data_type ioctlv_data;

struct _ioctl_data_type{
   int ioctl;
   unsigned char buffer_in[0x100];
   int len_in;
   unsigned char buffer_io[0x100];
   int len_io;
};

typedef struct _ioctl_data_type _ioctl_data_type;

_ioctl_data_type ioctl_data;

struct _seek_data_type{
   int where;
   int whence;
};

typedef struct _seek_data_type _seek_data_type;

_seek_data_type seek_data;

struct _write_data_type{

   unsigned char data[20];
   int len;
};

typedef struct _write_data_type _write_data_type;

_write_data_type write_data;

typedef unsigned char byte ;

typedef union byteLongConversion
{
   byte buffer[ sizeof(long) ] ;
   unsigned long key ;
} ByteConversionLong;

typedef union byteShortConversion
{
   byte buffer[ sizeof(short) ] ;
   unsigned short key ;
} ByteConversionShort;

void build_header(char* operation)
{
#ifdef LINUX_TEST
   char* new_line = "\r\n";
#else
   char* new_line = "\n";
#endif

   //strcpy(current_header,"");
   sprintf(current_header, "%s v%s %sBy Beardface%sByte Input Mode: (%s) Hit B to swap%s%s%s%s%s", 
         "EasyIOS",version,new_line,
         new_line,
         input_mode[current_input_mode],new_line,
         operation,new_line,
         "================================================",new_line);
}

#ifdef LINUX_TEST
int key = 0;

int get_Char()
{
   int rkey = 0;
   while((rkey = getchar()) == 0){}

   if(rkey == 93){return KEY_DOWN;}
   else if(rkey == 91){return KEY_UP;}
   else if(rkey == 13){return KEY_ENTER;}
   else if(rkey == 116){return MODE_CHANGE;}
   else if(rkey ==100){return KEY_RIGHT;}
   else if(rkey == 97){return KEY_LEFT;}
   else if(rkey == 119){return KEY_UP;}
   else if(rkey == 115){return KEY_DOWN;}
   else if(rkey == 101){return KEY_JUMP_R;}
   else if(rkey == 113){return KEY_JUMP_L;}

   return rkey;
}
#endif

void clearScreen()
{
#ifndef LINUX_TEST
   //VIDEO_ClearFrameBuffer(rmode,xfb,BLACK_COLOR);
   console_init(xfb,20,20,rmode->fbWidth,rmode->xfbHeight,rmode->fbWidth*VI_DISPLAY_PIX_SZ);
   VIDEO_WaitVSync();
#else
   system("clear");
#endif
}

int getInt()
{
   int val;
   int done = 0;
   static char* int_menu_options[10] = {
      "0",
      "1",
      "2",
      "3",
      "4",
      "5",
      "6",
      "7",
      "8",
      "9"};

   _menu_type int_menu = construct_menu(int_menu_options, 10);

#ifndef LINUX_TEST
   int buttonsDown = 0;
#endif
   strcpy(tfoot, "");
   build_header("Choose Int Val...");
   clearScreen();
   draw_menu_and_log(current_header, tfoot, &int_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);


   while(done == 0)
   {
#ifndef LINUX_TEST
      PAD_ScanPads();

      buttonsDown = PAD_ButtonsDown(0);
#else
      key = get_Char();
#endif

#ifndef LINUX_TEST
      if(buttonsDown & PAD_BUTTON_UP)
#else
         if(key == KEY_UP)
#endif
         {
            up_menu_item(&int_menu);
            clearScreen();
            draw_menu_and_log(current_header, tfoot, &int_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
         }
#ifndef LINUX_TEST
         else if(buttonsDown & PAD_BUTTON_DOWN)
#else
         else if(key == KEY_DOWN)
#endif
         {
            down_menu_item(&int_menu);
            clearScreen();
            draw_menu_and_log(current_header, tfoot, &int_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
         }
#ifndef LINUX_TEST
         else if((buttonsDown & PAD_BUTTON_A) || (buttonsDown & PAD_BUTTON_START))
#else
         else if(key == KEY_ENTER)
#endif
         {
            val = atoi(selected_menu_text(&int_menu));
            done = 1;
         }
   }

   return val;
}


int getOpenMode()
{
   int done = 0;
   static char* open_menu_options[3] = {
      "Read",
      "Write",
      "Cancel"};

   _menu_type open_menu = construct_menu(open_menu_options, 3);
#ifndef LINUX_TEST
   int buttonsDown = 0;
#endif
   strcpy(tfoot, "");

   build_header("Choose IOS Open Mode...");
   clearScreen();
   draw_menu_and_log(current_header, tfoot, &open_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);

   while(done == 0)
   {
#ifndef LINUX_TEST
      PAD_ScanPads();

      buttonsDown = PAD_ButtonsDown(0);
#else
      key = get_Char();
#endif

#ifndef LINUX_TEST
      if(buttonsDown & PAD_BUTTON_UP)
#else
         if(key == KEY_UP)
#endif
         {
            up_menu_item(&open_menu);
            clearScreen();
            draw_menu_and_log(current_header, tfoot, &open_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
         }
#ifndef LINUX_TEST
         else if(buttonsDown & PAD_BUTTON_DOWN)
#else
         else if(key == KEY_DOWN)
#endif
         {
            down_menu_item(&open_menu);
            clearScreen();
            draw_menu_and_log(current_header, tfoot, &open_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
         }
#ifndef LINUX_TEST
         else if((buttonsDown & PAD_BUTTON_A) || (buttonsDown & PAD_BUTTON_START))
#else
         else if(key == KEY_ENTER)
#endif
         {
            switch(open_menu.current_selected_index)
            {
               case 0:
                  return 1;
               case 1:
                  return 2;
               default:
                  return CANCEL;
            }
            done = 1;
         }
   }

   return CANCEL;
}

char getNibble(int byte_number, int nibble_number)
{
   char nibble = '0';
   int done = 0;
   static char* nibble_menu_options[16] = {
      "0",
      "1",
      "2",
      "3",
      "4",
      "5",
      "6",
      "7",
      "8",
      "9",
      "A",
      "B",
      "C",
      "D",
      "E",
      "F"};

   _menu_type nibble_menu = construct_menu(nibble_menu_options, 16);

#ifndef LINUX_TEST
   int buttonsDown = 0;
#endif
   build_header("Choose Nibble...");
   strcpy(tfoot,"");
   sprintf(tfoot, "Choosing nibble for byte:%i Nibble%i\r\n", byte_number, nibble_number);

   clearScreen();
   draw_menu_and_log(current_header, tfoot, &nibble_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);

   while(done == 0)
   {
#ifndef LINUX_TEST
      PAD_ScanPads();

      buttonsDown = PAD_ButtonsDown(0);
#else
      key = get_Char();
#endif

#ifndef LINUX_TEST
      if(buttonsDown & PAD_BUTTON_UP)
#else
         if(key == KEY_UP)
#endif
         {
            up_menu_item(&nibble_menu);
            clearScreen();
            draw_menu_and_log(current_header, tfoot, &nibble_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
         }
#ifndef LINUX_TEST
         else if(buttonsDown & PAD_BUTTON_DOWN)
#else
         else if(key == KEY_DOWN)
#endif
         {
            down_menu_item(&nibble_menu);
            clearScreen();
            draw_menu_and_log(current_header, tfoot, &nibble_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
         }
#ifndef LINUX_TEST
         else if((buttonsDown & PAD_BUTTON_A) || (buttonsDown & PAD_BUTTON_START))
#else
         else if(key == KEY_ENTER)
#endif
         {
            nibble = selected_menu_text(&nibble_menu)[0];
            done = 1;
         }
   }

   return nibble;
}

void update_bit_menu(_menu_type* menu, int index, int bits[])
{
   static char str_bits[10];
   static char str2_bits[10];
   static char nib1[5];
   static char nib2[5];

   static char str_ptr[10];
   static char str_hex[10];

   unsigned short tmp;

   strcpy(str_ptr, "          ");

   if(index > 3)
      str_ptr[index+1] = '^';
   else
      str_ptr[index] = '^';

   strcpy(nib1,"     ");
   sprintf(nib1,"%i%i%i%i",
         bits[0],bits[1],bits[2],bits[3]);

   strcpy(nib2,"     ");
   sprintf(nib2,"%i%i%i%i",
         bits[4],bits[5],bits[6],bits[7]);

   strcpy(str_bits, "          ");
   sprintf(str_bits, "%s|%s",
         nib1,nib2);

   strcpy(str2_bits, "          ");
   sprintf(str2_bits, "%s%s",nib1,nib2);

   tmp = strtoul(str2_bits, NULL, 2);

   strcpy(str_hex, "          ");
   sprintf(str_hex, "%.2X", tmp);
   menu->menu_items[0] = str_bits;
   menu->menu_items[1] = str_ptr;
   menu->menu_items[2] = str_hex;
}

char getBinaryByte()
{
   unsigned long val;
   int done = 0;
   int t_index = 0;
   char retByte = 0x00;
   int bits[8] = {0,0,0,0,0,0,0,0};

   char retval[10];

   char* bit_menu_options[3] = {
      "0000|0000",
      "^        ",
      "00"};

   _menu_type bit_menu = construct_menu(bit_menu_options, 3);
   
#ifndef LINUX_TEST
   int buttonsDown = 0;
#endif
   strcpy(tfoot,"");
   build_header("(A/Up/Down: flip, Left/Right: Step, R/L: Jump, Start: Done");
   clearScreen();
   draw_menu_and_log(current_header, tfoot, &bit_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
   
   while(done == 0)
   {
#ifndef LINUX_TEST
      PAD_ScanPads();

      buttonsDown = PAD_ButtonsDown(0);
#else
      key = get_Char();
#endif

#ifndef LINUX_TEST
      if(buttonsDown & PAD_BUTTON_UP || buttonsDown & PAD_BUTTON_DOWN || buttonsDown & PAD_BUTTON_A)
#else
         if(key == KEY_UP || key == KEY_DOWN)
#endif
         {
            if(bits[t_index] == 0){ bits[t_index] = 1;}
            else{bits[t_index] = 0;}

            update_bit_menu(&bit_menu, t_index, bits);
            clearScreen();
            draw_menu_and_log(current_header, tfoot, &bit_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
         }
#ifndef LINUX_TEST
         else if(buttonsDown & PAD_BUTTON_RIGHT)
#else
         else if(key == KEY_RIGHT)
#endif
         {
            if(t_index < 7)
            {
               t_index = t_index + 1;
               update_bit_menu(&bit_menu, t_index, bits);   
               clearScreen();
               draw_menu_and_log(current_header, tfoot, &bit_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
            }
         }
#ifndef LINUX_TEST
         else if(buttonsDown & PAD_BUTTON_LEFT)
#else
         else if(key == KEY_LEFT)
#endif
         {
            if(t_index > 0)
            {
               t_index = t_index - 1;
               update_bit_menu(&bit_menu, t_index, bits);
               clearScreen();
               draw_menu_and_log(current_header, tfoot, &bit_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
            }
         }
#ifndef LINUX_TEST
         else if(buttonsDown & PAD_TRIGGER_R)
#else
         else if(key == KEY_JUMP_R)
#endif
         {
               t_index = 7;
               update_bit_menu(&bit_menu, t_index, bits);
               clearScreen();
               draw_menu_and_log(current_header, tfoot, &bit_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
         }
#ifndef LINUX_TEST
         else if(buttonsDown & PAD_TRIGGER_L)
#else
         else if(key == KEY_JUMP_L)
#endif
         {
               t_index = 0;
               update_bit_menu(&bit_menu, t_index, bits);
               clearScreen();
               draw_menu_and_log(current_header, tfoot, &bit_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
         }

#ifndef LINUX_TEST
         else if((buttonsDown & PAD_BUTTON_START))
#else
         else if(key == KEY_ENTER)
#endif
         {
            ByteConversionShort bsc;

            strcpy(retval,"");
            sprintf(retval,"%i%i%i%i%i%i%i%i",
                  bits[0],bits[1],bits[2],bits[3],
                  bits[4],bits[5],bits[6],bits[7]);
            val = strtoul(retval, NULL, 2);

            bsc.key = (short)val;

#ifdef LINUX_TEST
            retByte = bsc.buffer[0];
#else
            retByte = bsc.buffer[sizeof(short)-1];
#endif
            done = 1;
         }
   }

   return retByte;

}

char getByte(int byte_number)
{
   static char my_byte[2];
   static unsigned long tmp;
   static char nib0;
   static char nib1;
   static char retByte = 0x00;

   if(current_input_mode == HEX_MODE)
   {
      ByteConversionShort bsc;

      nib0 = getNibble(byte_number, 0);
      nib1 = getNibble(byte_number, 1);
      sprintf(my_byte, "%c%c", nib0, nib1);

      tmp = strtoul(my_byte, NULL, 16);

      bsc.key = (short)tmp;

#ifdef LINUX_TEST
      retByte = bsc.buffer[0];
#else
      retByte = bsc.buffer[sizeof(short)-1];
#endif

   }
   else
   {
      retByte = getBinaryByte();
   }


   return retByte;
}

unsigned short getWord()
{
   ByteConversionShort bsc;

   bsc.buffer[0] = getByte(0);
   bsc.buffer[1] = getByte(1);


   printf("Word:%i\r\n", bsc.key);

   return bsc.key;
}

int main(int argc, char **argv) {
   unsigned char* key_press = 0x00;

   int app_done = 0;
   int command_done = 0;

   enum MENU_STATE {
      MAIN,
      IOS_OPERATION,
      IOS_WRITE,
      IOS_SEEK,
      IOCTL,
      IOCTLV
   };

   enum MENU_STATE state = MAIN;

   int j = 0;

   char selectedDevice[100] = "";
   int com_result = 0;

#ifndef LINUX_TEST
   int buttonsDown = 0;
#endif

   int fd;
   char footer[100] = "";
   char* main_operation = "Choose Interface...";

   char* main_menu_options[33] = {
      "/dev/aes",
      "/dev/boot2",
      "/dev/di",
      "/dev/es",
      "/dev/flash",
      "/dev/fs",
      "/dev/listen",
      "/dev/net/ip/bottom",
      "/dev/net/ip/top",
      "/dev/net/kd/request",
      "/dev/net/kd/time",
      "/dev/net/ncd/manage",
      "/dev/net/ssl",
      "/dev/net/usbeth/top",
      "/dev/net/wd/command",
      "/dev/net/wd/top",
      "/dev/printserver",
      "/dev/sdio",
      "/dev/sdio/slot0",
      "/dev/sdio/slot1",
      "/dev/sha",
      "/dev/stm/eventhook",
      "/dev/stm/immediate",
      "/dev/usb",
      "/dev/usb/ehc",
      "/dev/usb/ehc/0b95/7720",
      "/dev/usb/kbd",
      "/dev/usb/oh0",
      "/dev/usb/oh0/0b95/7720",
      "/dev/usb/oh1",
      "/dev/usb/oh1/57e/305",
      "/dev/wl0",
      "Quit"};

   _menu_type main_menu = construct_menu(main_menu_options, 33);

   char* sub_operation = "Choose Operation...";

   char* sub_menu_options[8] = {
      "IOS_Open",  
      "IOS_Read",  
      "IOS_Write", 
      "IOS_Seek",  
      "iosctl",    
      "iosctlv",   
      "IOS_Close", 
      "Back (and Close Interface if Open)"}; 

   _menu_type sub_menu = construct_menu(sub_menu_options, 8);

   char* data_operation = "Set Data then Choose Send...";

   char* ioctlv_menu_options[6] = {
      "Set ioctl",
      "Set cnt_in",
      "Set cnt_io",
      "Add byte to payload",
      "Send",
      "Cancel"};

   _menu_type ioctlv_menu = construct_menu(ioctlv_menu_options, 6);

   char* ioctl_menu_options[5] = {
      "Set ioctl",
      "Add byte to buffer_in payload",
      "Add byte to buffer_io payload",
      "Send",
      "Cancel"};

   _menu_type ioctl_menu = construct_menu(ioctl_menu_options, 5);

   char* ios_write_menu_options[3] = {
      "add byte to payload",
      "Send",
      "Cancel"};

   _menu_type ios_write_menu = construct_menu(ios_write_menu_options, 3);

   char* ios_seek_menu_options[4] = {
      "Set Where",
      "Set Whence",
      "Send",
      "Cancel"};

   _menu_type ios_seek_menu = construct_menu(ios_seek_menu_options, 4);

   _menu_type* current_menu = &main_menu;
   build_header(main_operation);

#ifdef LINUX_TEST
   initscr();
   nl();
#endif

#ifndef LINUX_TEST  
   VIDEO_Init();
   PAD_Init();

   switch(VIDEO_GetCurrentTvMode()) {
      case VI_NTSC:
         rmode = &TVNtsc480IntDf;
         break;
      case VI_PAL:
         rmode = &TVPal528IntDf;
         break;
      case VI_MPAL:
         rmode = &TVMpal480IntDf;
         break;
      default:
         rmode = &TVNtsc480IntDf;
         break;
   }

   xfb = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
   console_init(xfb,20,20,rmode->fbWidth,rmode->xfbHeight,rmode->fbWidth*VI_DISPLAY_PIX_SZ);

   VIDEO_Configure(rmode);
   VIDEO_SetNextFramebuffer(xfb);
   VIDEO_SetBlack(FALSE);
   VIDEO_Flush();
   VIDEO_WaitVSync();
   if(rmode->viTVMode&VI_NON_INTERLACE) VIDEO_WaitVSync();
#endif

   my_log = construct_log();
   key_press = 0x00;
   fd = -1;

   while(app_done == 0)
   {
      clearScreen();
      draw_menu_and_log(current_header, footer, current_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);

      command_done = 0;

      while(command_done == 0)
      {
         if(1)
         {
#ifndef LINUX_TEST
            PAD_ScanPads();

            buttonsDown = PAD_ButtonsDown(0);
#else
            key = get_Char();
#endif

#ifndef LINUX_TEST
            if(buttonsDown & PAD_BUTTON_UP)
#else
               if(key == KEY_UP)
#endif
               {
                  up_menu_item(current_menu);
                  clearScreen();
                  draw_menu_and_log(current_header, footer, current_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);


               }
#ifndef LINUX_TEST
               else if(buttonsDown & PAD_BUTTON_DOWN)
#else
               else if(key == KEY_DOWN)
#endif
               {
                  down_menu_item(current_menu);
                  clearScreen();
                  draw_menu_and_log(current_header, footer, current_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
               }
#ifndef LINUX_TEST
               else if(buttonsDown & PAD_BUTTON_B)
#else
               else if(key == MODE_CHANGE)
#endif
               {
                  if(current_input_mode == HEX_MODE)
                  {
                     current_input_mode = BINARY_MODE;
                  }
                  else
                  {
                     current_input_mode = HEX_MODE;
                  }

                  if(current_menu == &main_menu)
                     build_header(main_operation);
                  else if(current_menu == &sub_menu)
                     build_header(sub_operation);
                  else
                     build_header(data_operation);

                  clearScreen();
                  draw_menu_and_log(current_header, footer, current_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
               }
#ifndef LINUX_TEST
               else if((buttonsDown & PAD_BUTTON_A) || (buttonsDown & PAD_BUTTON_START))
#else
               else if(key == KEY_ENTER)
#endif
               {
                  logmsg(&my_log, selected_menu_text(current_menu));
                  switch(state)
                  {
                     case MAIN:
                        if(current_menu->current_selected_index == 32) //Quit
                        {
                           app_done = 1;
                        }
                        else
                        {
                           sprintf(selectedDevice, "%s", selected_menu_text(current_menu));
                           build_header(sub_operation);
                           current_menu   = &sub_menu;
                           state = IOS_OPERATION;
                        }
                        break;
                     case IOS_OPERATION:
                        switch(current_menu->current_selected_index)
                        {
                           case 0:
                              //IOS Open
                              {
                                 if(fd < 0)
                                 {
                                    int mode = getOpenMode();
                                    if(mode != CANCEL)
                                    {
                                       strcpy(tResult,"");
                                       sprintf(tResult,"Running IOS_Open (%s, %i)... ", selectedDevice, mode);
                                       logmsg(&my_log, tResult);
                                       {
                                          
                                          strcpy(sLog,"");
                                          sprintf(sLog,"   IOS_Open (%s, %i) !", selectedDevice, mode);
                                          logmsg(&my_log, sLog);
                                       }
                                       clearScreen();
                                       draw_menu_and_log(current_header, footer, current_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
#ifndef LINUX_TEST
                                       fd = IOS_Open(selectedDevice, mode);
#else
                                       fd = 1;
#endif
                                       if(fd >= 0)
                                       {
                                          logmsg(&my_log, "Success!");
                                       }
                                       else
                                       {
                                          strcpy(buffer,"");
                                          sprintf(buffer,"IOS_Open retunred invalid fd (%i)... ", fd);
                                          logmsg(&my_log, buffer);
                                       }
                                    }
                                 }
                                 else
                                 {
                                    logmsg(&my_log, "Cannot run IOS Open until you close the open interface.");
                                 }

                              }
                              break;
                           case 1:
                              //IOS Read
                              if(fd >= 0)
                              {
                                 int result = 0;
                                 logmsg(&my_log, "Running IOS_Read");
                                 strcpy(buffer,"");
                                 {
                                    
                                    strcpy(sLog,"");
                                    sprintf(sLog,"   IOS_Read (%i,&buffer[0x100],0x100)", fd);
                                    logmsg(&my_log, sLog);
                                 }
                                 clearScreen();
                                 draw_menu_and_log(current_header, footer, current_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
#ifndef LINUX_TEST
                                 result = IOS_Read(fd,&buffer,0x100);
#else
                                 result = 1;
#endif
                                 if(result == 0)
                                 {
                                    logmsg(&my_log, "IOS_Read Buffer:");
                                    logmsg(&my_log, buffer);
                                 }
                                 else
                                 {
                                    
                                    strcpy(tResult,"");
                                    sprintf(tResult,"IOS_Read Failed (%i) !", result);
                                    logmsg(&my_log, tResult);
                                 }
                              }
                              else
                              {
                                 logmsg(&my_log, "Can't run IOS Read until you open the device.");
                              }
                              break;
                           case 2:
                              //IOS Write
                              if(fd >= 0)
                              {
                                 logmsg(&my_log, "Running IOS_Write");
                                 state = IOS_WRITE;
                                 current_menu   = &ios_write_menu;
                              }
                              else
                              {
                                 logmsg(&my_log, "Can't run IOS Write until you open the device.");
                              }
                              break;
                           case 3:
                              //IOS Seek
                              if(fd >= 0)
                              {
                                 logmsg(&my_log, "Running IOS_Seek");
                                 state = IOS_SEEK;
                                 current_menu   = &ios_seek_menu;
                              }
                              else
                              {
                                 logmsg(&my_log, "Can't run IOS Seek until you open the device.");
                              }
                              break;
                           case 4:
                              //ioctl
                              if(fd >= 0)
                              {
                                 logmsg(&my_log, "Running ioctl");
                                 build_header(data_operation);
                                 current_menu   = &ioctl_menu;
                                 state = IOCTL;
                                 sprintf(footer,"IOS_Ioctl(%i,%i,{%s},%.2X, {%s}, %.2X)",
                                       fd, ioctl_data.ioctl, "", ioctl_data.len_in, "", ioctl_data.len_io);
                              }
                              else
                              {
                                 logmsg(&my_log, "Can't run ioctlv until you open the device.");
                              }
                              break;
                           case 5:
                              //ioctlv
                              if(fd >= 0)
                              {
                                 logmsg(&my_log, "Running ioctlv");
                                 build_header(data_operation); 
                                 current_menu   = &ioctlv_menu;
                                 state = IOCTLV;
                                 sprintf(footer,"IOS_Ioctlv(%i,%i,%i,%i,<{%s}>) payload length: %i",
                                       fd, ioctlv_data.ioctl, ioctlv_data.cnt_in, ioctlv_data.cnt_io,
                                       "", ioctlv_data.len);
                              }
                              else
                              {
                                 logmsg(&my_log, "Can't run ioctlv until you open the device.");
                              }
                              break;
                           case 6:
                              //IOS Close
                              if(fd >= 0)
                              {
                                 logmsg(&my_log, "Running IOS_Close");
                                 {
                                    
                                    strcpy(sLog,"");
                                    sprintf(sLog,"  IOS_Close (%i)", fd);
                                    logmsg(&my_log, sLog);
                                 }
                                 clearScreen();
                                 draw_menu_and_log(current_header, footer, current_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
#ifndef LINUX_TEST
                                 IOS_Close(fd);
#endif
                                 logmsg(&my_log, "Closed.");
                                 fd = -1;
                              }
                              else
                              {
                                 logmsg(&my_log, "Can't run IOS Close until you open the device.");
                                 fd = -1;
                              }
                              break;
                           case 7:
                              //Back
                              if(fd >= 0)
                              {
                                 logmsg(&my_log, "Running IOS_Close");
                                 {
                                    
                                    strcpy(sLog,"");
                                    sprintf(sLog,"  IOS_Close (%i)", fd);
                                    logmsg(&my_log, sLog);
                                 }
                                 clearScreen();
                                 draw_menu_and_log(current_header, footer, current_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
#ifndef LINUX_TEST
                                 IOS_Close(fd);
#endif
                                 logmsg(&my_log, "Closed.");
                                 fd = -1;
                              }
                              build_header(main_operation);
                              current_menu   = &main_menu;
                              state = MAIN; 
                              break;
                           default:
                              break;
                        }
                        break;
                     case IOCTLV:
                        //sprintf(footer,"%s","Building IOCTL Message...");
                        //display normal menu b/c not waiting on data

                        switch(current_menu->current_selected_index)
                        {
                           case 0: //ioctl
                              ioctlv_data.ioctl = getInt();
                              build_header(data_operation);
                              break;
                           case 1: //cnt_in
                              ioctlv_data.cnt_in = getInt();
                              build_header(data_operation);
                              break;
                           case 2: //cnt_io
                              ioctlv_data.cnt_io = getInt();
                              build_header(data_operation);
                              break;
                           case 3: //add byte to payload
                              {
                                 char myByte = getByte(0);
                                 ioctlv_data.data[ioctlv_data.len] = myByte;
                                 ioctlv_data.len = ioctlv_data.len + 1;
                                 build_header(data_operation);
                              }
                              break;
                           case 4: //Send!
                              logmsg(&my_log, "Sending ioctlv message...");
                              state = IOS_OPERATION;

#ifndef LINUX_TEST

                              ioctlv ioctl_args;
                              ioctl_args.data = &(ioctlv_data.data);
                              ioctl_args.len = ioctlv_data.len;
#endif

                              {//Log IOCTL Operation
                                 
                                 

                                 strcpy(payload,"");
                                 strcpy(tstr,"");
                                 for(j = 0; j < ioctlv_data.len; j++)
                                 {
                                    sprintf(tstr,"%.2X",ioctlv_data.data[j]);
                                    strcat(payload, tstr);
                                    if(j < (ioctlv_data.len - 1))
                                       strcat(payload, ":");
                                 }

                                 strcpy(sLog,"");
                                 sprintf(sLog, "  IOS_Ioctlv(%i,%i,%i,%i,<{%s}, %i>)",fd,ioctlv_data.ioctl,ioctlv_data.cnt_in, ioctlv_data.cnt_io,payload,ioctlv_data.len);
                                 logmsg(&my_log, sLog);
                              }           
                              clearScreen();
                              draw_menu_and_log(current_header, footer, current_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);

#ifndef LINUX_TEST
                              com_result = IOS_Ioctlv(fd, ioctlv_data.ioctl,  ioctlv_data.cnt_in, ioctlv_data.cnt_io, &ioctl_args);
#else
                              com_result = 1;
#endif
                              {
                                 strcpy(tResult,"");
                                 sprintf(tResult,"IOS_Ioctlv returned (%i)", com_result);
                                 logmsg(&my_log, tResult);
                              }
                              //reinit
                              ioctlv_data.len = 0;
                              sprintf(footer," ");
                              build_header(sub_operation);
                              current_menu   = &sub_menu;
                              break;
                           case 5:
                              //Cancel!
                              logmsg(&my_log, "Cancelled IOS_Ioctlv message.");
                              state = IOS_OPERATION;
                              ioctlv_data.len = 0;
                              sprintf(footer," ");
                              build_header(sub_operation);
                              current_menu   = &sub_menu;
                              break;
                        }

                        if(state == IOCTLV)
                        {
                           
                           strcpy(payload,"");
                           strcpy(tstr,"");
                           for(j = 0; j < ioctlv_data.len; j++)
                           {
                              sprintf(tstr,"%.2X",ioctlv_data.data[j]);
                              strcat(payload, tstr);
                              if(j < (ioctlv_data.len - 1))
                                 strcat(payload, ":");
                           }
                           strcpy(footer,"");
                           sprintf(footer,"IOS_Ioctlv(%i,%i,%i,%i,<{%s}>) payload length: %i",
                                 fd, ioctlv_data.ioctl, ioctlv_data.cnt_in, ioctlv_data.cnt_io,
                                 payload, ioctlv_data.len);
                        }

                        break;
                     case IOCTL:
                        switch(current_menu->current_selected_index)
                        {
                           case 0:
                              ioctl_data.ioctl = getInt();
                              build_header(data_operation);
                              break;
                           case 1: //add byte to buffer_in payload
                              {
                                 char myByte = getByte(0);
                                 ioctl_data.buffer_in[ioctl_data.len_in] = myByte;
                                 ioctl_data.len_in = ioctl_data.len_in + 1;
                                 build_header(data_operation);
                              }
                              break;
                           case 2: //add byte to buffer_in payload
                              {
                                 char myByte = getByte(0);
                                 ioctl_data.buffer_io[ioctl_data.len_io] = myByte;
                                 ioctl_data.len_io = ioctl_data.len_io + 1;
                                 build_header(data_operation);
                              }
                              break;
                              //Send!
                           case 3:
                              logmsg(&my_log, "Sending IOCTL message...");
                              state = IOS_OPERATION;
                              // reinit

                              {//Log IOCTL Operation

                                 strcpy(payload,"");
                                 strcpy(tstr,"");
                                 for(j = 0; j < ioctl_data.len_in; j++)
                                 {
                                    sprintf(tstr,"%.2X",ioctl_data.buffer_in[j]);
                                    strcat(payload, tstr);
                                    if(j < (ioctl_data.len_in - 1))
                                       strcat(payload, ":");
                                 }

                                 strcpy(buffer,"");
                                 strcpy(tstr,"");
                                 for(j = 0; j < ioctl_data.len_io; j++)
                                 {
                                    sprintf(tstr,"%.2X",ioctl_data.buffer_io[j]);
                                    strcat(buffer, tstr);
                                    if(j < (ioctl_data.len_io - 1))
                                       strcat(buffer, ":");
                                 }

                                 strcpy(sLog,"");
                                 sprintf(sLog, "  IOS_Ioctl(%i,%i,{%s},%i,{%s},%i)",fd,ioctl_data.ioctl,payload,ioctl_data.len_in,buffer,ioctl_data.len_io);
                                 logmsg(&my_log, sLog);
                              }
                              clearScreen();
                              draw_menu_and_log(current_header, footer, current_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
#ifndef LINUX_TEST
                              com_result = IOS_Ioctl(fd, ioctl_data.ioctl, &(ioctl_data.buffer_in), ioctl_data.len_in, &(ioctl_data.buffer_io), ioctl_data.len_io);
#else
                              com_result = 1;
#endif
                              {
                                 
                                 strcpy(tResult,"");
                                 sprintf(tResult,"IOS_Ioctl returned (%i) ", com_result);
                                 logmsg(&my_log, tResult);
                              }

                              ioctl_data.len_io = 0;
                              ioctl_data.len_in = 0;
                              sprintf(footer," ");
                              build_header(sub_operation);
                              current_menu   = &sub_menu;
                              break;
                           case 4:
                              //Cancel!
                              logmsg(&my_log, "Cancelled IOCTL message.");
                              state = IOS_OPERATION;
                              ioctl_data.len_io = 0;
                              ioctl_data.len_in = 0;
                              sprintf(footer," ");
                              build_header(sub_operation);
                              current_menu   = &sub_menu;
                              break;
                        }

                        if(state == IOCTL)
                        {

                           strcpy(payload,"");
                           strcpy(tstr,"");
                           for(j = 0; j < ioctl_data.len_in; j++)
                           {
                              sprintf(tstr,"%.2X",ioctl_data.buffer_in[j]);
                              strcat(payload, tstr);
                              if(j < (ioctl_data.len_in - 1))
                                 strcat(payload, ":");
                           }

                           strcpy(buffer,"");
                           strcpy(tstr,"");
                           for(j = 0; j < ioctl_data.len_io; j++)
                           {
                              sprintf(tstr,"%.2X",ioctl_data.buffer_io[j]);
                              strcat(buffer, tstr);
                              if(j < (ioctl_data.len_io - 1))
                                 strcat(buffer, ":");
                           }

                           strcpy(footer,"");
                           sprintf(footer,"IOS_Ioctl(%i,%i,{%s},%.2X, {%s}, %.2X)",
                                 fd, ioctl_data.ioctl, payload, ioctl_data.len_in, buffer, ioctl_data.len_io);
                        }

                        break;
                     case IOS_WRITE:
                        switch(current_menu->current_selected_index)
                        {
                           case 0: //add byte to payload
                              {
                                 char myByte = getByte(0);
                                 write_data.data[write_data.len] = myByte;
                                 write_data.len = write_data.len + 1;
                                 build_header(data_operation);
                              }
                              break;
                           case 1: //Send!
                              logmsg(&my_log, "Sending IOS_Write message...");
                              state = IOS_OPERATION;
                              // reinit
                              {//Log Write Operation
                                 
                                 strcpy(tstr,"");
                                 strcpy(payload,"");
                                 for(j = 0; j < write_data.len; j++)
                                 {
                                    sprintf(tstr,"%.2X",write_data.data[j]);
                                    strcat(payload, tstr);
                                    if(j < (write_data.len - 1))
                                       strcat(payload, ":");
                                 }
                                 strcpy(sLog,"");
                                 sprintf(sLog, "  IOS_Write(%i,{%s},%i)",fd,payload,write_data.len);
                                 logmsg(&my_log, sLog);
                              }
                              clearScreen();
                              draw_menu_and_log(current_header, footer, current_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
#ifndef LINUX_TEST
                              com_result = IOS_Write(fd, &(write_data.data), write_data.len);
#else
                              com_result = 1;
#endif
                              {
                                 
                                 strcpy(tResult,"");
                                 sprintf(tResult,"IOS_Write returned (%i)", com_result);
                                 logmsg(&my_log, tResult);
                              }

                              write_data.len = 0;
                              sprintf(footer," ");
                              build_header(sub_operation);
                              current_menu   = &sub_menu;
                              break;
                           case 2:
                              //Cancel!
                              logmsg(&my_log, "Cancelled IOS_Write message.");
                              state = IOS_OPERATION;
                              write_data.len = 0;
                              sprintf(footer," ");
                              build_header(sub_operation);
                              current_menu   = &sub_menu;
                              break;
                        }

                        if(state == IOS_WRITE)
                        {
                           
                           strcpy(payload,"");
                           strcpy(tstr,"");
                           for(j = 0; j < write_data.len; j++)
                           {
                              sprintf(tstr,"%.2X",write_data.data[j]);
                              strcat(payload, tstr);
                              if(j < (write_data.len - 1))
                                 strcat(payload, ":");
                           }
                           strcpy(footer,"");
                           sprintf(footer,"IOS_Write(%i,{%s},%.2X)",
                                 fd, payload, ioctlv_data.len);
                        }

                        break;
                     case IOS_SEEK:
                        switch(current_menu->current_selected_index)
                        {
                           case 0: 
                              seek_data.where = getInt();
                              build_header(data_operation);
                              break;
                           case 1:
                              seek_data.whence = getInt();
                              build_header(data_operation);
                              break;
                           case 2: //Send!
                              logmsg(&my_log, "Sending IOS_Seek message...");
                              state = IOS_OPERATION;
                              // reinit
                              {//Log Seek Operation
                                 
                                 strcpy(sLog,"");
                                 sprintf(sLog, "  IOS_Seek(%i,%i,%i)",fd,seek_data.where,seek_data.whence);
                                 logmsg(&my_log, sLog);
                              }
                              clearScreen();
                              draw_menu_and_log(current_header, footer, current_menu, MENU_SCREEN_SIZE, &my_log, LOG_SCREEN_SIZE);
#ifndef LINUX_TEST
                              com_result = IOS_Seek(fd, seek_data.where, seek_data.whence);
#else
                              com_result = 1;
#endif
                              {
                                 
                                 strcpy(tResult,"");
                                 sprintf(tResult,"IOS_Seek Failed (%i) !", com_result);
                                 logmsg(&my_log, tResult);
                              }

                              sprintf(footer," ");
                              build_header(sub_operation);
                              current_menu   = &sub_menu;
                              break;
                           case 3:
                              //Cancel!
                              logmsg(&my_log, "Cancelled IOS_Seek message.");
                              state = IOS_OPERATION;
                              sprintf(footer," ");
                              build_header(sub_operation);
                              current_menu   = &sub_menu;
                              break;
                        }

                        if(state == IOS_SEEK)
                        {
                           sprintf(footer,"IOS_Seek(%i,%i,%i)",
                                 fd, seek_data.where, seek_data.whence);
                        }

                        break;
                     default:
                        state = IOS_OPERATION;
                        current_menu = &sub_menu;
                        break;
                  }
                  command_done = 1;
               }
         }
      }
   }

#ifdef LINUX_TEST 
   endwin();
   exit(0);
#endif
   return 0;
}
